/*
 * Copyright (C) 2006 mark307
 * See Copyright Notice in COPYING.LUA
 */

#include <windows.h>
#include <string.h>
#include <lua.h>
#include <lauxlib.h>

#include "array.h"
#include "list.h"
#include "dialog.h"
#include "prop.h"

#define LIBNAME "w32dlg"


/**
   
   * Win32 Tiny dialog functions
   
   
   ** Defined variables
   
   WS_CHILD WS_VISIBLE WS_DISABLED WS_CAPTION WS_BORDER WS_DLGFRAME
   WS_VSCROLL WS_HSCROLL WS_SYSMENU WS_THICKFRAME WS_GROUP WS_TABSTOP
   WS_MINIMIZEBOX WS_MAXIMIZEBOX WS_TILED WS_SIZEBOX WS_TILEDWINDOW
   WS_OVERLAPPEDWINDOW WS_POPUPWINDOW WS_CHILDWINDOW
   
   WS_EX_DLGMODALFRAME WS_EX_TOPMOST WS_EX_TRANSPARENT WS_EX_TOOLWINDOW
   WS_EX_WINDOWEDGE WS_EX_CLIENTEDGE WS_EX_RIGHT WS_EX_LEFT WS_EX_RTLREADING
   WS_EX_LTRREADING WS_EX_LEFTSCROLLBAR WS_EX_RIGHTSCROLLBAR
   WS_EX_CONTROLPARENT WS_EX_STATICEDGE WS_EX_OVERLAPPEDWINDOW
   WS_EX_PALETTEWINDOW
   
   DS_ABSALIGN DS_SETFONT DS_MODALFRAME DS_NOIDLEMSG DS_SETFOREGROUND
   DS_3DLOOK DS_FIXEDSYS DS_CENTER DS_CENTERMOUSE
   
   ES_LEFT ES_CENTER ES_RIGHT ES_MULTILINE ES_UPPERCASE ES_LOWERCASE
   ES_PASSWORD ES_AUTOVSCROLL ES_AUTOHSCROLL ES_NOHIDESEL ES_OEMCONVERT
   ES_READONLY ES_WANTRETURN ES_NUMBER
   
   BS_PUSHBUTTON BS_DEFPUSHBUTTON BS_CHECKBOX BS_AUTOCHECKBOX
   BS_RADIOBUTTON BS_3STATE BS_AUTO3STATE BS_GROUPBOX BS_AUTORADIOBUTTON
   BS_LEFTTEXT BS_TEXT BS_LEFT BS_RIGHT BS_CENTER BS_TOP BS_BOTTOM
   BS_VCENTER BS_PUSHLIKE BS_MULTILINE BS_NOTIFY BS_FLAT BS_RIGHTBUTTON
   BST_UNCHECKED BST_CHECKED BST_INDETERMINATE BST_PUSHED BST_FOCUS
   
   SS_LEFT SS_CENTER SS_RIGHT SS_BLACKRECT SS_GRAYRECT SS_WHITERECT
   SS_BLACKFRAME SS_GRAYFRAME SS_WHITEFRAME SS_SIMPLE SS_LEFTNOWORDWRAP
   SS_ETCHEDHORZ SS_ETCHEDVERT SS_ETCHEDFRAME SS_NOPREFIX SS_NOTIFY
   SS_RIGHTJUST SS_SUNKEN SS_ENDELLIPSIS SS_PATHELLIPSIS SS_WORDELLIPSIS
   SS_ELLIPSISMASK
   
   LBS_NOTIFY LBS_SORT LBS_MULTIPLESEL LBS_HASSTRINGS LBS_USETABSTOPS
   LBS_NOINTEGRALHEIGHT LBS_MULTICOLUMN LBS_EXTENDEDSEL
   LBS_DISABLENOSCROLL LBS_STANDARD
   
   CBS_SIMPLE CBS_DROPDOWN CBS_DROPDOWNLIST CBS_AUTOHSCROLL CBS_OEMCONVERT
   CBS_SORT CBS_NOINTEGRALHEIGHT CBS_DISABLENOSCROLL CBS_UPPERCASE
   CBS_LOWERCASE
   
   IDOK IDCANCEL
   
   
   ** Defined functions
   
*/


/*
   matatable functions
 */

static int L_gc(lua_State *L)
{
	DialogTpl *d;
	
	d = luaL_checkudata(L, 1, LIBNAME);
	DialogTpl_Free(d);
	lua_settop(L, 0);
	return 0;
}

/**
   
   *** functions
   
 */

/**
   PropertySheet
   
   Show property sheet.
   
      result = w32dlg.PropertySheet(dlg,
                                    [hwnd, [hinstance, [caption, [sel]]]])
   
   Show property sheet contains dialog object and return.
   If error will occur, it returns nil.
   
   dlg       : table    : table contains w32dlg objects
   hwnd      : integer  : window handle
   hinstance : integer  : hinstance
   caption   : string   : caption
   sel       : integer  : select page index
   
   result    : integer  : id or nil
                          OK >0, Cancel 0
 */
static int L_PropertySheet(lua_State *L)
{
	DialogTpl **dialogs, *d;
	const char *caption;
	HWND hwnd;
	HANDLE hinst;
	int i, n, sel, result;
	
	luaL_checktype(L, 1, LUA_TTABLE);
	hwnd = (HWND) luaL_optinteger(L, 2, (lua_Integer) NULL);
	hinst = (HANDLE) luaL_optinteger(L, 3, (lua_Integer) NULL);
	caption = luaL_optstring(L, 4, NULL);
	sel = luaL_optinteger(L, 5, 0);
	if (lua_gettop(L) > 4) lua_settop(L, 4);
	
	n = lua_objlen(L, 1);
	dialogs = lua_newuserdata(L, sizeof(DialogTpl *) * n);
	for (i = 0; i < n; i ++) {
		lua_rawgeti(L, 1, i + 1);
		d = luaL_checkudata(L, -1, LIBNAME);
		dialogs[i] = d;
		// no pop
	}
	
	result = PropertySheet_ShowModal(hwnd, hinst, caption, sel, -n, dialogs);
	if (result < 0) {
		return 0;
	}
	lua_pushinteger(L, result);
	return 1;
}


/**
   
   *** constructor
   
 */

/**
   New
   
   Create dialog object.
   
      result = w32dlg.New(style, exstyle, x, y, cx, cy
                          [, caption [, font [, point]]])
   
   Create dialog object and return. If error will occur, it returns nil.
   
   style   : integer : window style
   exstyle : integer : extended window style
   x       : integer : x position
   y       : integer : y position
   cx      : integer : window width
   cy      : integer : window height
   caption : string  : window caption
   font    : string  : font
   point   : integer : font size
   
   result  : userdata : w32dlg object
 */
static int L_New(lua_State *L)
{
	DialogTpl *d;
	DWORD style, dwExtendedStyle;
	short x, y, cx, cy;
	LPCTSTR caption;
	LPCTSTR font;
	WORD point;
	int result;
	
	style = luaL_checkinteger(L, 1);
	dwExtendedStyle = luaL_checkinteger(L, 2);
	x = luaL_checkinteger(L, 3);
	y = luaL_checkinteger(L, 4);
	cx = luaL_checkinteger(L, 5);
	cy = luaL_checkinteger(L, 6);
	caption = luaL_optstring(L, 7, "");
	font = luaL_optstring(L, 8, NULL);
	point = luaL_optinteger(L, 9, 0);
	
	d = lua_newuserdata(L, sizeof(DialogTpl));
	result = DialogTpl_New(d, style, dwExtendedStyle, x, y, cx, cy,
		caption, font, point);
	if (result) {
		return 0;
	}
	luaL_getmetatable(L, LIBNAME);
	lua_setmetatable(L, -2);
	return 1;
}


/**
   
   ** method
   
 */

/**
   GetButtonResult 
   
   Get result as number from control with id.
   
      result = dlg:GetButtonResult(id)
   
   dlg     : userdata : w32dlg
   id      : integer  : control id
   
   result  : integer  : result
 */
static int L_GetNumber(lua_State *L)
{
	DialogTpl *d;
	int id, n, result;
	ListType *l;
	
	d = luaL_checkudata(L, 1, LIBNAME);
	id = luaL_checkinteger(L, 2);
	
	l = DialogTpl_GetResult(d, id);
	if (!l) return 0;
	result = ListType_GetNumber(l, &n);
	if (result) return 0;
	lua_pushinteger(L, n);
	return 1;
}

/**
   SetButtonInitial
   
   Set initial number to control with id.
   
      result = dlg:SetButtonInitial(id, value)
   
   dlg     : userdata : w32dlg
   id      : integer  : control id
   value   : integer  : initial number
   
   result  : true or nil : success or fail
 */
static int L_SetNumber(lua_State *L)
{
	DialogTpl *d;
	int id, n, result;
	ListType *l;
	
	d = luaL_checkudata(L, 1, LIBNAME);
	id = luaL_checkinteger(L, 2);
	n = luaL_checkinteger(L, 3);
	
	l = DialogTpl_GetInitialData(d, id);
	if (!l) return 0;
	result = ListType_SetNumber(l, n);
	if (result) return 0;
	lua_pushboolean(L, 1);
	return 1;
}

/**
   GetEditResult GetComboboxResult
   
   Get result as string from control with id.
   
      result = dlg:GetEditResult(id)
      result = dlg:GetComboboxResult(id)
   
   dlg     : userdata : w32dlg
   id      : integer  : control id
   
   result  : string   : result
 */
static int L_GetString(lua_State *L)
{
	DialogTpl *d;
	int id, result;
	ListType *l;
	char *ptr = NULL;
	
	d = luaL_checkudata(L, 1, LIBNAME);
	id = luaL_checkinteger(L, 2);
	
	l = DialogTpl_GetResult(d, id);
	if (!l) return 0;
	result = ListType_GetBuffer(l, &ptr);
	if (result || ! ptr) return 0;
	lua_pushstring(L, ptr);
	return 1;
}

/**
   SetEditInitial SetStaticInitial
   

   Set initial string to control with id.
   
      result = dlg:SetEditInitial(id, value)
      result = dlg:SetStaticInitial(id, value)
   
   dlg     : userdata : w32dlg
   id      : integer  : control id
   value   : string   : initial string
   
   result  : true or nil : success or fail
 */
static int L_SetString(lua_State *L)
{
	DialogTpl *d;
	int id, result;
	ListType *l;
	const char *ptr;
	size_t n;
	
	d = luaL_checkudata(L, 1, LIBNAME);
	id = luaL_checkinteger(L, 2);
	ptr = luaL_checkstring(L, 3);
	n = lua_objlen(L, 3) + 1;
	
	l = DialogTpl_GetInitialData(d, id);
	if (!l) return 0;
	result = ListType_CopyBuffer(l, ptr, n);
	if (result) return 0;
	lua_pushboolean(L, 1);
	return 1;
}

/**
   GetListboxResult
   
   Get resulting from list control with id.
   
      result = dlg:GetListboxResult(id)
   
   dlg     : userdata : w32dlg
   id      : integer  : control id
   
   result  : table    : integer array contains selected index
 */
static int L_GetListboxResult(lua_State *L)
{
	DialogTpl *d;
	int id, result;
	ListType *l;
	int *ptr = NULL;
	int i, n;
	
	d = luaL_checkudata(L, 1, LIBNAME);
	id = luaL_checkinteger(L, 2);
	
	l = DialogTpl_GetResult(d, id);
	if (!l) return 0;
	lua_newtable(L);
	switch (l->type) {
	case NUMBER:
		result = ListType_GetNumber(l, &n);
		if (result) return 0;
		lua_pushinteger(L, n + 1);
		lua_rawseti(L, -2, 1);
		break;
	case BUFFER:
		result = ListType_GetBuffer(l, &ptr);
		if (result || ! ptr) return 0;
		for (i = 1; *ptr >= 0; ptr ++, i ++) {
			lua_pushinteger(L, *ptr + 1);
			lua_rawseti(L, -2, i);
		}
		break;
	case LIST:
		return 0;
	}
	return 1;
}

static int set_strings(lua_State *L, int index, ListType *l)
{
	const char *ptr;
	size_t n;
	int result;
	
	lua_pushnil(L);
	while (lua_next(L, index)) {
		ptr = lua_tostring(L, -1);
		if (ptr) {
			n = lua_objlen(L, -1) + 1;
			result = ListType_PushCopyBuffer(l, ptr, n);
			if (result) return 1;
		}
		lua_pop(L, 1);
	}
	return 0;
}

/**
   SetListboxInitial
   
   Set items and initial selection to listbox control with id.
   
      result = dlg:SetListboxInitial(id, initial, items)
   
   dlg     : userdata : w32dlg
   id      : integer  : control id
   initial : table    : integer array contains indexes to select
   items   : table    : string array contains items
   
   result  : true or nil : success or fail
 */
static int L_SetListboxInitial(lua_State *L)
{
	DialogTpl *d;
	int id, result;
	ListType *l;
	int *ptr = NULL;
	size_t i, n, size;
	
	d = luaL_checkudata(L, 1, LIBNAME);
	id = luaL_checkinteger(L, 2);
	luaL_checktype(L, 3, LUA_TTABLE);
	luaL_checktype(L, 4, LUA_TTABLE);
	lua_settop(L, 4);
	
	l = DialogTpl_GetInitialData(d, id);
	if (!l) return 0;
	ListType_Finalize(l);
	result = set_strings(L, 4, l);
	if (result) return 0;
	result = ListType_Reverse(l);
	if (result) return 0;
	
	n = lua_objlen(L, 3);
	size = sizeof(int) * (n + 1);
	ptr = lua_newuserdata(L, size);
	for (i = 0; i < n; i ++) {
		lua_rawgeti(L, 3, i + 1);
		if (lua_isnumber(L, -1)) {
			ptr[i] = lua_tointeger(L, -1) - 1;
		}
		lua_pop(L, 1);
	}
	ptr[n] = -1;
	result = ListType_PushCopyBuffer(l, ptr, size);
	if (result) return 0;
	lua_pushboolean(L, 1);
	return 1;
}

/**
   SetComboboxInitial
   
   Set items and initial selection to combobox control with id.
   
      result = dlg:SetListboxInitial(id, initial, items)
   
   dlg     : userdata : w32dlg
   id      : integer  : control id
   initial : string   : initial string
   items   : table    : string array contains items
   
   result  : true or nil : success or fail
 */
static int L_SetComboboxInitial(lua_State *L)
{
	DialogTpl *d;
	int id, result;
	ListType *l;
	const char *ptr = NULL;
	size_t n;
	
	d = luaL_checkudata(L, 1, LIBNAME);
	id = luaL_checkinteger(L, 2);
	ptr = luaL_checkstring(L, 3);
	luaL_checktype(L, 4, LUA_TTABLE);
	lua_settop(L, 4);
	
	n = lua_objlen(L, 3) + 1;
	
	l = DialogTpl_GetInitialData(d, id);
	if (!l) return 0;
	ListType_Finalize(l);
	result = set_strings(L, 4, l);
	if (result) return 0;
	result = ListType_Reverse(l);
	if (result) return 0;
	result = ListType_PushCopyBuffer(l, ptr, n);
	if (result) return 0;
	lua_pushboolean(L, 1);
	return 1;
}

/**
   AddButton AddStatic AddEdit AddListbox AddCombobox
   
   Add control to dialog.
   
      result = dlg:AddButton(style, exstyle, x, y, cx, cy, id
                             [, caption])
      result = dlg:AddStatic(style, exstyle, x, y, cx, cy, id
                             [, caption])
      result = dlg:AddEdit(style, exstyle, x, y, cx, cy, id
                           [, caption])
      result = dlg:AddListbox(style, exstyle, x, y, cx, cy, id
                              [, caption])
      result = dlg:AddCombobox(style, exstyle, x, y, cx, cy, id
                               [, caption])
   
   style   : integer : window style
   exstyle : integer : extended window style
   x       : integer : x position
   y       : integer : y position
   cx      : integer : control width
   cy      : integer : control height
   id      : integer : control id
   caption : string  : caption
   
   result  : true or nil : success or fail
 */
static int AddControl(lua_State *L,
	int (*f)(DialogTpl *, DWORD, DWORD,
		short, short, short, short, WORD, LPCTSTR))
{
	DialogTpl *d;
	DWORD style, dwExtendedStyle;
	short x, y, cx, cy;
	LPCTSTR caption;
	WORD id;
	int result;
	
	d = luaL_checkudata(L, 1, LIBNAME);
	style = luaL_checkinteger(L, 2);
	dwExtendedStyle = luaL_checkinteger(L, 3);
	x = luaL_checkinteger(L, 4);
	y = luaL_checkinteger(L, 5);
	cx = luaL_checkinteger(L, 6);
	cy = luaL_checkinteger(L, 7);
	id = luaL_checkinteger(L, 8);
	caption = luaL_optstring(L, 9, "");
	
	result = (*f)(d, style, dwExtendedStyle, x, y, cx, cy, id, caption);
	if (result) return 0;
	lua_pushboolean(L, 1);
	return 1;
}

static int L_AddButton(lua_State *L)
{
	return AddControl(L, DialogTpl_AddButton);
}
static int L_AddStatic(lua_State *L)
{
	return AddControl(L, DialogTpl_AddStatic);
}
static int L_AddEdit(lua_State *L)
{
	return AddControl(L, DialogTpl_AddEdit);
}
static int L_AddListbox(lua_State *L)
{
	return AddControl(L, DialogTpl_AddListbox);
}
static int L_AddCombobox(lua_State *L)
{
	return AddControl(L, DialogTpl_AddCombobox);
}


/**
   ReleaseResult
   
   Release memory contains results.
   
      dlg:ReleaseResults()
 */
static int L_ReleaseResult(lua_State *L)
{
	DialogTpl *dt;
	DialogItem *di;
	int i;
	
	dt = luaL_checkudata(L, 1, LIBNAME);
	di = (DialogItem *) dt->item.ptr;
	if (di) {
		for (i = 0; i < dt->num; i ++, di ++) {
			ListType_Finalize(&di->result);
		}
	}
	return 0;
}


/**
   ShowModal
   
   Show dialog as modal.
   
      result = dlg:ShowModal([caption [, hinstance [, hwnd]]])
   
   dlg       : userdata : w32dlg
   caption   : string   : caption
   hinstance : integer  : hinstance
   hwnd      : integer  : window handle
   
   result    : integer  : id
 */
static int L_ShowModal(lua_State *L)
{
	DialogTpl *d;
	const char *caption;
	HWND hwnd;
	HANDLE hinst;
	int result;
	
	d = luaL_checkudata(L, 1, LIBNAME);
	caption = luaL_optstring(L, 2, NULL);
	hinst = (HANDLE) luaL_optinteger(L, 3, (lua_Integer) NULL);
	hwnd = (HWND) luaL_optinteger(L, 4, (lua_Integer) NULL);
	
	result = DialogTpl_ShowModal(d, caption, hinst, hwnd);
	lua_pushinteger(L, result);
	return 1;
}


static struct {
	const char *name;
	DWORD value;
} consts[] = {
	/* */
	{"WS_POPUP",            WS_POPUP},
	{"WS_CHILD",            WS_CHILD},
	{"WS_VISIBLE",          WS_VISIBLE},
	{"WS_DISABLED",         WS_DISABLED},
	{"WS_CAPTION",          WS_CAPTION},
	{"WS_BORDER",           WS_BORDER},
	{"WS_DLGFRAME",         WS_DLGFRAME},
	{"WS_VSCROLL",          WS_VSCROLL},
	{"WS_HSCROLL",          WS_HSCROLL},
	{"WS_SYSMENU",          WS_SYSMENU},
	{"WS_THICKFRAME",       WS_THICKFRAME},
	{"WS_GROUP",            WS_GROUP},
	{"WS_TABSTOP",          WS_TABSTOP},
	{"WS_MINIMIZEBOX",      WS_MINIMIZEBOX},
	{"WS_MAXIMIZEBOX",      WS_MAXIMIZEBOX},
	{"WS_TILED",            WS_TILED},
	{"WS_SIZEBOX",          WS_SIZEBOX},
	{"WS_TILEDWINDOW",      WS_TILEDWINDOW},
	{"WS_OVERLAPPEDWINDOW", WS_OVERLAPPEDWINDOW},
	{"WS_POPUPWINDOW",      WS_POPUPWINDOW},
	{"WS_CHILDWINDOW",      WS_CHILDWINDOW},
	/* */
	{"WS_EX_DLGMODALFRAME",     WS_EX_DLGMODALFRAME},
	{"WS_EX_TOPMOST",           WS_EX_TOPMOST},
	{"WS_EX_TRANSPARENT",       WS_EX_TRANSPARENT},
	{"WS_EX_TOOLWINDOW",        WS_EX_TOOLWINDOW},
	{"WS_EX_WINDOWEDGE",        WS_EX_WINDOWEDGE},
	{"WS_EX_CLIENTEDGE",        WS_EX_CLIENTEDGE},
	{"WS_EX_RIGHT",             WS_EX_RIGHT},
	{"WS_EX_LEFT",              WS_EX_LEFT},
	{"WS_EX_RTLREADING",        WS_EX_RTLREADING},
	{"WS_EX_LTRREADING",        WS_EX_LTRREADING},
	{"WS_EX_LEFTSCROLLBAR",     WS_EX_LEFTSCROLLBAR},
	{"WS_EX_RIGHTSCROLLBAR",    WS_EX_RIGHTSCROLLBAR},
	{"WS_EX_CONTROLPARENT",     WS_EX_CONTROLPARENT},
	{"WS_EX_STATICEDGE",        WS_EX_STATICEDGE},
	{"WS_EX_OVERLAPPEDWINDOW",  WS_EX_OVERLAPPEDWINDOW},
	{"WS_EX_PALETTEWINDOW",     WS_EX_PALETTEWINDOW},
	/* */
	{"DS_ABSALIGN",         DS_ABSALIGN},
	{"DS_SETFONT",          DS_SETFONT},
	{"DS_MODALFRAME",       DS_MODALFRAME},
	{"DS_NOIDLEMSG",        DS_NOIDLEMSG},
	{"DS_SETFOREGROUND",    DS_SETFOREGROUND},
	{"DS_3DLOOK",           DS_3DLOOK},
	{"DS_FIXEDSYS",         DS_FIXEDSYS},
	{"DS_CENTER",           DS_CENTER},
	{"DS_CENTERMOUSE",      DS_CENTERMOUSE},
	/* */
	{"ES_LEFT",             ES_LEFT},
	{"ES_CENTER",           ES_CENTER},
	{"ES_RIGHT",            ES_RIGHT},
	{"ES_MULTILINE",        ES_MULTILINE},
	{"ES_UPPERCASE",        ES_UPPERCASE},
	{"ES_LOWERCASE",        ES_LOWERCASE},
	{"ES_PASSWORD",         ES_PASSWORD},
	{"ES_AUTOVSCROLL",      ES_AUTOVSCROLL},
	{"ES_AUTOHSCROLL",      ES_AUTOHSCROLL},
	{"ES_NOHIDESEL",        ES_NOHIDESEL},
	{"ES_OEMCONVERT",       ES_OEMCONVERT},
	{"ES_READONLY",         ES_READONLY},
	{"ES_WANTRETURN",       ES_WANTRETURN},
	{"ES_NUMBER",           ES_NUMBER},
	/* */
	{"BS_PUSHBUTTON",       BS_PUSHBUTTON},
	{"BS_DEFPUSHBUTTON",    BS_DEFPUSHBUTTON},
	{"BS_CHECKBOX",         BS_CHECKBOX},
	{"BS_AUTOCHECKBOX",     BS_AUTOCHECKBOX},
	{"BS_RADIOBUTTON",      BS_RADIOBUTTON},
	{"BS_3STATE",           BS_3STATE},
	{"BS_AUTO3STATE",       BS_AUTO3STATE},
	{"BS_GROUPBOX",         BS_GROUPBOX},
	{"BS_AUTORADIOBUTTON",  BS_AUTORADIOBUTTON},
	{"BS_LEFTTEXT",         BS_LEFTTEXT},
	{"BS_TEXT",             BS_TEXT},
	{"BS_LEFT",             BS_LEFT},
	{"BS_RIGHT",            BS_RIGHT},
	{"BS_CENTER",           BS_CENTER},
	{"BS_TOP",              BS_TOP},
	{"BS_BOTTOM",           BS_BOTTOM},
	{"BS_VCENTER",          BS_VCENTER},
	{"BS_PUSHLIKE",         BS_PUSHLIKE},
	{"BS_MULTILINE",        BS_MULTILINE},
	{"BS_NOTIFY",           BS_NOTIFY},
	{"BS_FLAT",             BS_FLAT},
	{"BS_RIGHTBUTTON",      BS_RIGHTBUTTON},
	/* */
	{"BST_UNCHECKED",       BST_UNCHECKED},
	{"BST_CHECKED",         BST_CHECKED},
	{"BST_INDETERMINATE",   BST_INDETERMINATE},
	{"BST_PUSHED",          BST_PUSHED},
	{"BST_FOCUS",           BST_FOCUS},
	/* */
	{"SS_LEFT",             SS_LEFT},
	{"SS_CENTER",           SS_CENTER},
	{"SS_RIGHT",            SS_RIGHT},
	{"SS_BLACKRECT",        SS_BLACKRECT},
	{"SS_GRAYRECT",         SS_GRAYRECT},
	{"SS_WHITERECT",        SS_WHITERECT},
	{"SS_BLACKFRAME",       SS_BLACKFRAME},
	{"SS_GRAYFRAME",        SS_GRAYFRAME},
	{"SS_WHITEFRAME",       SS_WHITEFRAME},
	{"SS_SIMPLE",           SS_SIMPLE},
	{"SS_LEFTNOWORDWRAP",   SS_LEFTNOWORDWRAP},
	{"SS_ETCHEDHORZ",       SS_ETCHEDHORZ},
	{"SS_ETCHEDVERT",       SS_ETCHEDVERT},
	{"SS_ETCHEDFRAME",      SS_ETCHEDFRAME},
	{"SS_NOPREFIX",         SS_NOPREFIX},
	{"SS_NOTIFY",           SS_NOTIFY},
	{"SS_RIGHTJUST",        SS_RIGHTJUST},
	{"SS_SUNKEN",           SS_SUNKEN},
	{"SS_ENDELLIPSIS",      SS_ENDELLIPSIS},
	{"SS_PATHELLIPSIS",     SS_PATHELLIPSIS},
	{"SS_WORDELLIPSIS",     SS_WORDELLIPSIS},
	{"SS_ELLIPSISMASK",     SS_ELLIPSISMASK},
	/* */
	{"LBS_NOTIFY",            LBS_NOTIFY},
	{"LBS_SORT",              LBS_SORT},
	{"LBS_MULTIPLESEL",       LBS_MULTIPLESEL},
	{"LBS_HASSTRINGS",        LBS_HASSTRINGS},
	{"LBS_USETABSTOPS",       LBS_USETABSTOPS},
	{"LBS_NOINTEGRALHEIGHT",  LBS_NOINTEGRALHEIGHT},
	{"LBS_MULTICOLUMN",       LBS_MULTICOLUMN},
	{"LBS_EXTENDEDSEL",       LBS_EXTENDEDSEL},
	{"LBS_DISABLENOSCROLL",   LBS_DISABLENOSCROLL},
	{"LBS_STANDARD",          LBS_STANDARD},
	/* */
	{"CBS_SIMPLE",            CBS_SIMPLE},
	{"CBS_DROPDOWN",          CBS_DROPDOWN},
	{"CBS_DROPDOWNLIST",      CBS_DROPDOWNLIST},
	{"CBS_AUTOHSCROLL",       CBS_AUTOHSCROLL},
	{"CBS_OEMCONVERT",        CBS_OEMCONVERT},
	{"CBS_SORT",              CBS_SORT},
	{"CBS_NOINTEGRALHEIGHT",  CBS_NOINTEGRALHEIGHT},
	{"CBS_DISABLENOSCROLL",   CBS_DISABLENOSCROLL},
	{"CBS_UPPERCASE",         CBS_UPPERCASE},
	{"CBS_LOWERCASE",         CBS_LOWERCASE},
	/* */
	{"IDOK",                  IDOK},
	{"IDCANCEL",              IDCANCEL},
	{NULL, 0},
};

static const luaL_Reg metas[] = {
	{"__gc", L_gc},
	{NULL, NULL},
};

static const luaL_Reg methods[] = {
	{"AddButton", L_AddButton},
	{"AddStatic", L_AddStatic},
	{"AddEdit", L_AddEdit},
	{"AddListbox", L_AddListbox},
	{"AddCombobox", L_AddCombobox},
	/* */
	{"GetButtonResult", L_GetNumber},
	{"SetButtonInitial", L_SetNumber},
	/* */
	{"GetEditResult", L_GetString},
	{"SetEditInitial", L_SetString},
	/* */
	{"SetStaticInitial", L_SetString},
	/* */
	{"GetListboxResult", L_GetListboxResult},
	{"SetListboxInitial", L_SetListboxInitial},
	/* */
	{"GetComboboxResult", L_GetString},
	{"SetComboboxInitial", L_SetComboboxInitial},
	/* */
	{"ReleaseResult", L_ReleaseResult},
	/* */
	{"ShowModal", L_ShowModal},
	{"Show", L_ShowModal},
	{NULL, NULL},
};

static const luaL_Reg funcs[] = {
	{"New", L_New},
	{"PropertySheet", L_PropertySheet},
	{NULL, NULL},
};

__declspec(dllexport) int luaopen_w32dlg(lua_State *L)
{
	int i;
	const char *name;
	
	name = luaL_optstring(L, 1, LIBNAME);
	if (*name == '\0') {
		name = LIBNAME;
	}
	
	/* define metatable */
	if (luaL_newmetatable(L, LIBNAME)) {
		luaL_register(L, NULL, metas);
		lua_newtable(L);
		luaL_register(L, NULL, methods);
		lua_setfield(L, -2, "__index");
	}
	lua_pop(L, 1);
	
	/* define module */
	luaL_register(L, name, funcs);
	for (i = 0; consts[i].name != NULL; i++) {
		lua_pushnumber(L, consts[i].value);
		lua_setfield(L, -2, consts[i].name);
	}
	return 1;
}
