/*
 * Copyright (C) 2006 mark307
 * See Copyright Notice in COPYING.LUA
 */

#include <windows.h>
#include <lua.h>
#include <lauxlib.h>

#define LIBNAME "w32mnu"


/**
   
   * Win32 Menu functions
   
   ** Defined variables
   
   MF_ENABLED MF_GRAYED MF_DISABLED MF_CHECKED MF_MENUBARBREAK MF_MENUBREAK
   MF_POPUP MF_SEPARATOR MF_STRING MF_DEFAULT MF_SYSMENU MF_HELP MF_END
   MF_RIGHTJUSTIFY MF_MOUSESELECT MF_INSERT MF_CHANGE MF_APPEND MF_DELETE
   MF_REMOVE MF_USECHECKBITMAPS MF_UNHILITE MF_HILITE MF_BYCOMMAND
   MF_BYPOSITION MF_UNCHECKED MF_HILITE MF_UNHILITE
   
   
   ** Defined functions
   
*/


struct menu {
	HMENU hmenu;
	LPVOID attached;
};

/*
   matatable functions
 */

static int L_gc(lua_State *L)
{
	struct menu *lpmenu;
	
	lpmenu = luaL_checkudata(L, 1, LIBNAME);
	if (lpmenu->hmenu && ! lpmenu->attached) {
		DestroyMenu(lpmenu->hmenu);
	}
	lua_settop(L, 0);
	return 0;
}

/**
   
   ** static method
   
*/

/**
   DrawMenuBar
   
   Wrapper of DrawMenuBar Win32 API.
   
      result = DrawMenuBar(hwnd)
   
   hwnd   : integer : Window handle
   result : boolean : 
 */
static int L_DrawMenuBar(lua_State *L)
{
	HWND hwnd = (HWND) luaL_checkinteger(L, 1);
	
	lua_pushboolean(L, DrawMenuBar(hwnd));
	return 1;
}


/**
   
   *** constructor
   
 */
static int L_new(lua_State *L, HMENU hmenu, LPVOID attached)
{
	struct menu *lpmenu;
	
	if (hmenu && IsMenu(hmenu)) {
		lpmenu = lua_newuserdata(L, sizeof(struct menu));
		lpmenu->hmenu = hmenu;
		lpmenu->attached = attached;
		luaL_getmetatable(L, LIBNAME);
		lua_setmetatable(L, -2);
		return 1;
	}
	return 0;
}

/**
   CreateMenu
   
   Wrapper of CreateMenu Win32 API.
   
      hmenu = CreateMenu()
   
   hmenu : userdata : w32mnu
*/
static int L_CreateMenu(lua_State *L)
{
	return L_new(L, CreateMenu(), NULL);
}

/**
   CreatePopupMenu
   
   Wrapper of CreatePopupMenu Win32 API.
   
      hmenu = CreatePopupMenu()
   
   hmenu : userdata : w32mnu
*/
static int L_CreatePopupMenu(lua_State *L)
{
	return L_new(L, CreatePopupMenu(), NULL);
}

/**
   GetMenu
   
   Wrapper of GetMenu Win32 API.
   
      hmenu = GetMenu(hwnd)
   
   hwnd  : integer  : Window Handle
   hmenu : userdata : w32mnu
*/
static int L_GetMenu(lua_State *L)
{
	HWND hwnd = (HWND) luaL_checkinteger(L, 1);
	
	if (hwnd && IsWindow(hwnd)) {
		return L_new(L, GetMenu(hwnd), hwnd);
	}
	return 0;
}

/**
   AttachMenu
   
   Attach w32mnu userdata object to menu handle.
   
      menu = AttachMenu(hmenu)
   
   hmenu  : integer  : Menu handle
   menu   : userdata : w32mnu
*/
static int L_AttachMenu(lua_State *L)
{
	HMENU hmenu = (HMENU) luaL_checkinteger(L, 1);
	
	if (hmenu) {
		return L_new(L, hmenu, hmenu);
	}
	return 0;
}


/**
   
   ** method
   
 */

/**
   CheckMenuItem
   
   Wrapper of CheckMenuItem Win32 API.
   
      result = menu:CheckMenuItem(id, flags)
   
   menu   : userdata : w32mnu
   id     : integer  :
   flags  : integer  :
   result : boolean  :
*/
static int L_CheckMenuItem(lua_State *L)
{
	struct menu *lpmenu = luaL_checkudata(L, 1, LIBNAME);
	UINT idCheck = luaL_checkinteger(L, 2);
	UINT uFlags = luaL_checkinteger(L, 3);
	
	lua_pushboolean(L, CheckMenuItem(lpmenu->hmenu, idCheck, uFlags));
	return 1;
}

/**
   CheckMenuRadioItem
   
   Wrapper of CheckMenuRadioItem Win32 API.
   
      result = menu:CheckMenuRadioItem(first, last, id, flags)
   
   menu   : userdata : w32mnu
   first  : integer  :
   last   : integer  :
   id     : integer  :
   flags  : integer  :
   result : boolean  :
*/
static int L_CheckMenuRadioItem(lua_State *L)
{
	struct menu *lpmenu = luaL_checkudata(L, 1, LIBNAME);
	UINT idFirst = luaL_checkinteger(L, 2);
	UINT idLast = luaL_checkinteger(L, 3);
	UINT idCheck = luaL_checkinteger(L, 4);
	UINT uFlags = luaL_checkinteger(L, 5);
	
	lua_pushboolean(L, CheckMenuRadioItem(lpmenu->hmenu, idFirst, idLast,
		idCheck, uFlags));
	return 1;
}

/**
   DeleteMenu
   
   Wrapper of DeleteMenu Win32 API.
   
      result = menu:DeleteMenu(item, flags)
   
   menu   : userdata : w32mnu
   item   : integer  :
   flags  : integer  :
   result : boolean  :
*/
static int L_DeleteMenu(lua_State *L)
{
	struct menu *lpmenu = luaL_checkudata(L, 1, LIBNAME);
	UINT uItem = luaL_checkinteger(L, 2);
	UINT flags = luaL_checkinteger(L, 3);
	
	lua_pushboolean(L, DeleteMenu(lpmenu->hmenu, uItem, flags));
	return 1;
}

/**
   DestroyMenu
   
   Wrapper of DestroyMenu Win32 API.
   
      result = menu:DestroyMenu()
   
   menu   : userdata : w32mnu
   result : boolean  :
*/
static int L_DestroyMenu(lua_State *L)
{
	struct menu *lpmenu = luaL_checkudata(L, 1, LIBNAME);
	BOOL result = FALSE;
	
	if (! lpmenu->attached) {
		result = DestroyMenu(lpmenu->hmenu);
		if (result) {
			lpmenu->hmenu = NULL;
			lpmenu->attached = NULL;
		}
	}
	lua_pushboolean(L, result);
	return 1;
}

/**
   EnableMenuItem
   
   Wrapper of EnableMenuItem Win32 API.
   
      result = menu:EnableMenuItem(item, flags)
   
   menu   : userdata : w32mnu
   item   : integer  :
   flags  : integer  :
   result : boolean  :
*/
static int L_EnableMenuItem(lua_State *L)
{
	struct menu *lpmenu = luaL_checkudata(L, 1, LIBNAME);
	UINT uItem = luaL_checkinteger(L, 2);
	UINT flags = luaL_checkinteger(L, 3);
	
	lua_pushboolean(L, EnableMenuItem(lpmenu->hmenu, uItem, flags));
	return 1;
}

/**
   GetMenuItemCount
   
   Wrapper of GetMenuItemCount Win32 API.
   
      result = menu:GetMenuItemCount()
   
   menu   : userdata : w32mnu
   result : integer  :
*/
static int L_GetMenuItemCount(lua_State *L)
{
	struct menu *lpmenu = luaL_checkudata(L, 1, LIBNAME);
	
	lua_pushinteger(L, GetMenuItemCount(lpmenu->hmenu));
	return 1;
}

/**
   GetMenuItemID
   
   Wrapper of GetMenuItemID Win32 API.
   
      result = menu:GetMenuItemID()
   
   menu   : userdata : w32mnu
   result : integer  :
*/
static int L_GetMenuItemID(lua_State *L)
{
	struct menu *lpmenu = luaL_checkudata(L, 1, LIBNAME);
	int pos = luaL_checkinteger(L, 2);
	
	lua_pushinteger(L, GetMenuItemID(lpmenu->hmenu, pos));
	return 1;
}

/**
   GetMenuState
   
   Wrapper of GetMenuState Win32 API.
   
      result = menu:GetMenuState(item, flags)
   
   menu   : userdata : w32mnu
   item   : integer  :
   flags  : integer  :
   result : integer  :
*/
static int L_GetMenuState(lua_State *L)
{
	struct menu *lpmenu = luaL_checkudata(L, 1, LIBNAME);
	UINT uItem = luaL_checkinteger(L, 2);
	UINT flags = luaL_checkinteger(L, 3);
	
	lua_pushinteger(L, GetMenuState(lpmenu->hmenu, uItem, flags));
	return 1;
}

/**
   GetMenuString
   
   Wrapper of GetMenuString Win32 API.
   
      result = menu:GetMenuString(item, flags, new_item)
   
   menu     : userdata           : w32mnu
   item     : integer            :
   flags    : integer            :
   result   : string             :
*/
static int L_GetMenuString(lua_State *L)
{
	struct menu *lpmenu = luaL_checkudata(L, 1, LIBNAME);
	UINT uItem = luaL_checkinteger(L, 2);
	UINT flags = luaL_checkinteger(L, 3);
	int length;
	LPSTR buffer;
	
	length = GetMenuString(lpmenu->hmenu, uItem, NULL, 0, flags);
	if (length == 0) return 0;
	length ++;
	buffer = lua_newuserdata(L, length);
	GetMenuString(lpmenu->hmenu, uItem, buffer, length, flags);
	lua_pushstring(L, buffer);
	return 1;
}

/**
   GetSubMenu
   
   Wrapper of GetSubMenu Win32 API.
   
      result = menu:GetSubMenu(item, flags)
   
   menu   : userdata : w32mnu
   item   : integer  :
   flags  : integer  :
   result : userdata : w32mnu
*/
static int L_GetSubMenu(lua_State *L)
{
	struct menu *lpmenu = luaL_checkudata(L, 1, LIBNAME);
	UINT uItem = luaL_checkinteger(L, 2);
	
	return L_new(L, GetSubMenu(lpmenu->hmenu, uItem), lpmenu->hmenu);
}

/**
   HiliteMenuItem
   
   Wrapper of HiliteMenuItem Win32 API.
   
      result = menu:HiliteMenuItem(item, flags)
   
   menu   : userdata : w32mnu
   item   : integer  :
   flags  : integer  :
   result : boolean  :
*/
static int L_HiliteMenuItem(lua_State *L)
{
	struct menu *lpmenu = luaL_checkudata(L, 1, LIBNAME);
	UINT uItem = luaL_checkinteger(L, 2);
	UINT flags = luaL_checkinteger(L, 3);
	HWND hwnd = lpmenu->attached;
	BOOL result = FALSE;
	
	if (hwnd && IsWindow(hwnd)) {
		result = HiliteMenuItem(hwnd, lpmenu->hmenu, uItem, flags);
	}
	lua_pushboolean(L, result);
	return 1;
}

/**
   InsertMenu
   
   Wrapper of InsertMenu Win32 API.
   
      result = menu:InsertMenu(item, flags, new_item, label)
   
   menu     : userdata           : w32mnu
   item     : integer            :
   flags    : integer            :
   new_item : integer | userdata :
   label    : string             :
   
   result   : boolean            :
*/
static int L_InsertMenu(lua_State *L)
{
	struct menu *lpmenu = luaL_checkudata(L, 1, LIBNAME);
	UINT uItem = luaL_checkinteger(L, 2);
	UINT flags = luaL_checkinteger(L, 3);
	LPCSTR lpNewItem = (LPCSTR) luaL_checkstring(L, 5);
	BOOL result;
	
	if (flags & MF_POPUP) {
		struct menu *lppop;
		lppop = luaL_checkudata(L, 4, LIBNAME);
		result = InsertMenu(lpmenu->hmenu, uItem, flags,
			(UINT) lppop->hmenu, lpNewItem);
		if (result) {
			lppop->attached = lpmenu->hmenu;
		}
	} else {
		UINT idNewItem = luaL_checkinteger(L, 4);
		result = InsertMenu(lpmenu->hmenu, uItem, flags, idNewItem, lpNewItem);
	}
	lua_pushboolean(L, result);
	return 1;
}

/**
   ModifyMenu
   
   Wrapper of ModifyMenu Win32 API.
   
      result = menu:ModifyMenu(item, flags, new_item, label)
   
   menu     : userdata           : w32mnu
   item     : integer            :
   flags    : integer            :
   new_item : integer | userdata :
   label    : string             :
   
   result   : boolean            :
*/
static int L_ModifyMenu(lua_State *L)
{
	struct menu *lpmenu = luaL_checkudata(L, 1, LIBNAME);
	UINT uItem = luaL_checkinteger(L, 2);
	UINT flags = luaL_checkinteger(L, 3);
	LPCSTR lpNewItem = (LPCSTR) luaL_checkstring(L, 5);
	BOOL result;
	
	if (flags & MF_POPUP) {
		struct menu *lppop;
		lppop = luaL_checkudata(L, 4, LIBNAME);
		result = ModifyMenu(lpmenu->hmenu, uItem, flags,
			(UINT) lppop->hmenu, lpNewItem);
		if (result) {
			lppop->attached = lpmenu->hmenu;
		}
	} else {
		UINT idNewItem = luaL_checkinteger(L, 4);
		result = ModifyMenu(lpmenu->hmenu, uItem, flags, idNewItem, lpNewItem);
	}
	lua_pushboolean(L, result);
	return 1;
}

/**
   SetMenu
   
   Wrapper of SetMenu Win32 API.
   
      result = menu:SetMenu(hwnd)
   
   menu     : userdata : w32mnu
   hwnd     : integer  :
   result   : boolean  :
*/
static int L_SetMenu(lua_State *L)
{
	struct menu *lpmenu = luaL_checkudata(L, 1, LIBNAME);
	HWND hwnd = (HWND) luaL_checkinteger(L, 2);
	HMENU hmenu;
	BOOL result = FALSE;
	
	if (hwnd && IsWindow(hwnd)) {
		hmenu = GetMenu(hwnd);
		if (hmenu) {
			DestroyMenu(hmenu);
		}
		result = SetMenu(hwnd, lpmenu->hmenu);
		if (result) {
			lpmenu->attached = hwnd;
		}
	}
	lua_pushboolean(L, result);
	return 1;
}

/**
   TrackPopupMenu
   
   Wrapper of TrackPopupMenu Win32 API.
   
      result = menu:TrackPopupMenu(flags, x, y, hwnd)
      result = menu:TrackPopupMenu(flags, hwnd)
   
   menu     : userdata : w32mnu
   flags    : integer  :
   x        : integer  :
   y        : integer  :
   hwnd     : integer  :
   result   : integer  :
*/
static int L_TrackPopupMenu(lua_State *L)
{
	struct menu *lpmenu;
	UINT flags;
	int x, y;
	HWND hwnd;
	int result;
	
	lpmenu = luaL_checkudata(L, 1, LIBNAME);
	flags = (UINT) luaL_checkinteger(L, 2);
	if (lua_gettop(L) == 3) {
		POINT pt;
		hwnd = (HWND) luaL_checkinteger(L, 3);
		if (GetCursorPos(&pt)) {
			x = pt.x;
			y = pt.y;
		} else {
			x = y = 0;
		}
	} else {
		luaL_checkany(L, 5);
		x = luaL_checkinteger(L, 3);
		y = luaL_checkinteger(L, 4);
		hwnd = (HWND) luaL_checkinteger(L, 5);
	}
	result = TrackPopupMenu(lpmenu->hmenu, flags, x, y, 0, hwnd, NULL);
	lua_pushinteger(L, result);
	return 1;
}

static struct {
	const char *name;
	DWORD value;
} consts[] = {
	{"MF_ENABLED",         MF_ENABLED},
	{"MF_GRAYED",          MF_GRAYED},
	{"MF_DISABLED",        MF_DISABLED},
	/* {"MF_BITMAP",          MF_BITMAP}, */
	{"MF_CHECKED",         MF_CHECKED},
	{"MF_MENUBARBREAK",    MF_MENUBARBREAK},
	{"MF_MENUBREAK",       MF_MENUBREAK},
	/* {"MF_OWNERDRAW",       MF_OWNERDRAW}, */
	{"MF_POPUP",           MF_POPUP},
	{"MF_SEPARATOR",       MF_SEPARATOR},
	{"MF_STRING",          MF_STRING},
	{"MF_DEFAULT",         MF_DEFAULT},
	{"MF_SYSMENU",         MF_SYSMENU},
	{"MF_HELP",            MF_HELP},
	{"MF_END",             MF_END},
	{"MF_RIGHTJUSTIFY",    MF_RIGHTJUSTIFY},
	{"MF_MOUSESELECT",     MF_MOUSESELECT},
	{"MF_INSERT",          MF_INSERT},
	{"MF_CHANGE",          MF_CHANGE},
	{"MF_APPEND",          MF_APPEND},
	{"MF_DELETE",          MF_DELETE},
	{"MF_REMOVE",          MF_REMOVE},
	{"MF_USECHECKBITMAPS", MF_USECHECKBITMAPS},
	{"MF_UNHILITE",        MF_UNHILITE},
	{"MF_HILITE",          MF_HILITE},
	{"MF_BYCOMMAND",       MF_BYCOMMAND},
	{"MF_BYPOSITION",      MF_BYPOSITION},
	{"MF_UNCHECKED",       MF_UNCHECKED},
	{"MF_HILITE",          MF_HILITE},
	{"MF_UNHILITE",        MF_UNHILITE},
	{"TPM_CENTERALIGN",    TPM_CENTERALIGN},
	{"TPM_LEFTALIGN",      TPM_LEFTALIGN},
	{"TPM_RIGHTALIGN",     TPM_RIGHTALIGN},
	{"TPM_LEFTBUTTON",     TPM_LEFTBUTTON},
	{"TPM_RIGHTBUTTON",    TPM_RIGHTBUTTON},
	{"TPM_HORIZONTAL",     TPM_HORIZONTAL},
	{"TPM_VERTICAL",       TPM_VERTICAL},
	{"TPM_TOPALIGN",       TPM_TOPALIGN},
	{"TPM_VCENTERALIGN",   TPM_VCENTERALIGN},
	{"TPM_BOTTOMALIGN",    TPM_BOTTOMALIGN},
	{"TPM_NONOTIFY",       TPM_NONOTIFY},
	{"TPM_RETURNCMD",      TPM_RETURNCMD},
	{NULL, 0},
};

static const luaL_Reg metas[] = {
	{"__gc", L_gc},
	{NULL, NULL},
};

static const luaL_Reg methods[] = {
	{"CheckMenuItem", L_CheckMenuItem},
	{"CheckMenuRadioItem", L_CheckMenuRadioItem},
	{"DeleteMenu", L_DeleteMenu},
	{"DestroyMenu", L_DestroyMenu},
	{"EnableMenuItem", L_EnableMenuItem},
	{"GetMenuItemCount", L_GetMenuItemCount},
	{"GetMenuItemID", L_GetMenuItemID},
	{"GetMenuState", L_GetMenuState},
	{"GetMenuString", L_GetMenuString},
	{"GetSubMenu", L_GetSubMenu},
	{"HiliteMenuItem", L_HiliteMenuItem},
	{"InsertMenu", L_InsertMenu},
	{"ModifyMenu", L_ModifyMenu},
	{"SetMenu", L_SetMenu},
	{"TrackPopupMenu", L_TrackPopupMenu},
	{NULL, NULL},
};

static const luaL_Reg funcs[] = {
	{"CreateMenu", L_CreateMenu},
	{"CreatePopupMenu", L_CreatePopupMenu},
	{"GetMenu", L_GetMenu},
	{"AttachMenu", L_AttachMenu},
	{"DrawMenuBar", L_DrawMenuBar},
	{NULL, NULL},
};

__declspec(dllexport) int luaopen_w32mnu(lua_State *L)
{
	int i;
	const char *name;
	
	name = luaL_optstring(L, 1, LIBNAME);
	if (*name == '\0') {
		name = LIBNAME;
	}
	
	/* define metatable */
	if (luaL_newmetatable(L, LIBNAME)) {
		luaL_register(L, NULL, metas);
		lua_newtable(L);
		luaL_register(L, NULL, methods);
		lua_setfield(L, -2, "__index");
	}
	lua_pop(L, 1);
	
	/* define module */
	luaL_register(L, name, funcs);
	for (i = 0; consts[i].name != NULL; i++) {
		lua_pushnumber(L, consts[i].value);
		lua_setfield(L, -2, consts[i].name);
	}
	return 1;
}
